/*
 * mod_gunzip.c: pipe documents through a "send-gunzipped" handler.
 *
 * 03 March 1999	initial release
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_request.h"
#include "http_protocol.h"
#include "http_log.h"
#include "http_main.h"
#include "util_script.h"
#include "zlib.h"

#ifndef BUFSIZ
/* should come from stdio.h */
#define BUFSIZ 4096
#endif

static int gunzip_handler(request_rec *r)
{
    FILE *f;
    int errstatus;
    request_rec *parent;
    gzFile zf;

    r->allowed |= (1 << M_GET);
    if (r->method_number != M_GET)
	return DECLINED;

    if (r->finfo.st_mode == 0) {
	ap_log_rerror(APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, r,
	    "File does not exist: %s",
	    (r->path_info
	    ? ap_pstrcat(r->pool, r->filename, r->path_info, NULL)
	    : r->filename));
	return HTTP_NOT_FOUND;
    }

    if (!(f = ap_pfopen(r->pool, r->filename, "r"))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
	    "file permissions deny server access: %s", r->filename);
	return HTTP_FORBIDDEN;
    }

#ifdef PLAIN
#else
    if (!(zf = gzdopen(fileno(f), "r"))) {
	ap_log_rerror(APLOG_MARK, APLOG_ERR, r,
	    "file permissions deny server access: %s", r->filename);
	return HTTP_FORBIDDEN;
    }
#endif

    if ((errstatus = ap_meets_conditions(r)) != OK)
	return errstatus;

    ap_send_http_header(r);
    if (!r->header_only) {
	static char buf[BUFSIZ];
	int nread, nwritten;

	while ((nread =
#ifdef PLAIN
	    fread(buf, 1, BUFSIZ, f)
#else
	    gzread(zf, buf, BUFSIZ)
#endif
	    ) > 0) {
	    ap_soft_timeout("send gunzip", r);

	    while (!r->connection->aborted && nread > 0) {
		nwritten = ap_rwrite(buf, nread, r);
		if (nwritten > 0) {
		    ap_reset_timeout(r);
		    nread -= nwritten;
		    if (nread > 0)
			/* we're lazy... */
			memmove(buf, &buf[nwritten], nread);
		} else if (nwritten < 0) {
		    if (r->connection->aborted)
			break;
		    else if (errno == EAGAIN)
			continue;
		    else {
			ap_log_rerror(APLOG_MARK, APLOG_INFO, r,
			    "client stopped connection before send completed");
			ap_bsetflag(r->connection->client, B_EOUT, 1);
			r->connection->aborted = 1;
			break;
		    }
		}
	    }
	    ap_kill_timeout(r);
	    if (r->sent_bodyct)
		ap_bgetopt(r->connection->client, BO_BYTECT, &r->bytes_sent);
	}
    }
#ifdef PLAIN
#else
    gzclose(zf);
#endif
    ap_pfclose(r->pool, f);

    return OK;
}

static const handler_rec gunzip_handlers[] =
{
    {"send-gunzipped", gunzip_handler},
    {NULL}
};

module MODULE_VAR_EXPORT gunzip_module =
{
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
    NULL,			/* dir config creater */
    NULL,			/* dir merger --- default is to override */
    NULL,			/* server config */
    NULL,			/* merge server config */
    NULL,			/* command table */
    gunzip_handlers,		/* handlers */
    NULL,			/* filename translation */
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
    NULL,			/* type_checker */
    NULL,			/* fixups */
    NULL,			/* logger */
    NULL,			/* header parser */
    NULL,			/* child_init */
    NULL,			/* child_exit */
    NULL			/* post read-request */
};
